//
// nettop 
//
// Copyright (c) 2000
//           Scott R Parish <sRp@srparish.net>, OR  97119-9201.
//       All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer as
//    the first lines of this file unmodified.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY Scott R Parish ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL Scott R Parish BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//
// TODO
// 
// * figure ports out better
// * show udp ports
// * show more about 802.3 
// * have a mode where you only see statistics from the last x time period
// * allow for collapsing and expanding of sub-types (like tcp/udp/etc)
// * move eth types/ip types/etc out to seperate file (user defineable)
// * get tcp/udp from /etc/services
// * remove segfaults on screen resize
// * add support for ipv6
// * add a quet mode which dumps percentages into a file after a given time
// * make key strokes (such as 'q') instant instead of delayed)
// * handle sigint better (currently segfaults sometimes)
// * think about removing the "24" limit
//

#define VERSION "0.2.0"

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <slang.h>
#include <pcap.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>
#ifdef sun
#include <netinet/in_systm.h>
#include <netinet/in.h>
#endif
#include <netinet/ip.h>
#define __FAVOR_BSD 1
#include <netinet/tcp.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <time.h>
#include <stdlib.h>

#include "node.h"

//Colors
#define total_c    1
#define ethernet_c 2
#define ip_c       3
#define tcp_c      4
#define udp_c      4
#define bar_c      5
#define header_c   6


//do all the grunt work to get into permiscuious mode
pcap_t *pcap_start(char *filterstr);

//print the stats out
void screen_update();
void ipprint(int last[]);
void tcpprint(int last[]);

//call functions that need to be ran every alarm, and reset the alarm
void alarmed();

//tie things down before quiting
void dieing();

//initialize ncurses
void ui_run();
void change_screen_size();
void ts_run();
void sprintf_nice(char *str, unsigned int data); 
void change_delay();
void show_help();
void change_sort_type();
void pause_display();
void clear_count();
void usage();

void sl_print_row(int pkt, int pktf, char *str1, int tot, int totf, char *str2,
                  int szpkt, char *type, int level, int last[]); 
void sl_header(); 
void sl_init(); 
void sl_check_kb();

void t_header();
void t_print_row(int pkt, int pktf, char *str1, int tot, int totf,  
                  char *str2, int szpkt, char *type, int level, int last[]); 

struct node ether={0,0,-1,0,0};
struct node ip={0,0,-1,0,0};
struct node tcp={0,0,-1,0,0};

// which network device to use
char *dev=0;
               
extern char *optarg;
extern int optind;

int delay=1000000; //how long to wait between screen updates
int runcount=0; //how many times to run

int start_time;

int y;
int maxx,maxy;


unsigned char sort_type='p';  //either (p)acket or (s)ize

unsigned char ui_type='s';  //(s)lang, (t)ext
int t_print_head=0;
int print_lines=1;


pcap_t *pd;

pthread_mutex_t mutex;

int main(int argc, char **argv) {
  int ch;
  pthread_t ui; //user interface
  char *filterstr = " ";

  while((ch = getopt(argc,argv,"n:i:d:f:spthl")) != -1) {
    switch(ch) {
      case 'i':
        dev = optarg; break;
      case 'd':
        delay=(1000000*atof(optarg)); break;
      case 'f':
        filterstr = optarg;           break;
      case 'h':
        t_print_head = 1;             break;
      case 'l':
        print_lines = !print_lines;   break;
      case 'n':
        runcount = atoi(optarg);      break;
      case 'p':
        sort_type = 'p';              break;
      case 's':
        sort_type = 's';              break;
      case 't':
        ui_type = 't'; 
        print_lines = !print_lines;   break;
      case '?':
      default:
        usage();
        exit(1);
    }
  }

  pd = pcap_start(filterstr);

  pthread_create(&ui,NULL,(void*)&ui_run,NULL);
  signal(SIGINT, dieing);

  ts_run();
  
  exit(0);
}

void ts_run() {
  struct node *t;
  char *next_pcap(int *len);
  char *ptr;
  struct pcap_pkthdr hdr;
  unsigned short eth_type;
  int ip_head_len;
  struct ip iphdr;
  struct tcphdr tcphdr;
  int port;

  while(1) {
    while ( (ptr = (char *) pcap_next(pd, &hdr)) == NULL) ;
    
    //ethernetpacket type
    memcpy(&eth_type,&ptr[12],2);
    eth_type=ntohs(eth_type);

    if(eth_type<0x05dd) {  //this is an 802.3 packet
       unsigned short type; 
       memcpy(&type,&ptr[18],2);
       type=ntohs(type);

       pthread_mutex_lock(&mutex);
       if((t=get(&ether,type))); 
       else 
	  t=new(&ether,type);
       t->count++;
       t->size+=hdr.len;
       pthread_mutex_unlock(&mutex);
    }

    else {  //this is a ethernet packet

      pthread_mutex_lock(&mutex);
      if((t=get(&ether,eth_type))); 
      else 
	 t=new(&ether,eth_type);
      t->count++;
      t->size+=hdr.len;
      pthread_mutex_unlock(&mutex);

      switch(eth_type) {
      // IPv4
      case 0x0800: 
	memcpy(&iphdr,ptr+14,sizeof(struct ip));
	ip_head_len=iphdr.ip_hl*4;
	
	//which ip protocol is this
	pthread_mutex_lock(&mutex);
	if((t=get(&ip,iphdr.ip_p)));
	else 
	   t=new(&ip,iphdr.ip_p);
	t->count++;
	t->size+=hdr.len;
	pthread_mutex_unlock(&mutex);
	
	if(iphdr.ip_p==6) {
	  memcpy(&tcphdr,ptr+14+ip_head_len,sizeof(struct tcphdr));
	  
	  //work on this
	  port=(ntohs(tcphdr.th_sport) < ntohs(tcphdr.th_dport)? 
		ntohs(tcphdr.th_sport) : ntohs(tcphdr.th_dport));
	  
	  pthread_mutex_lock(&mutex);
	  if((t=get(&tcp,port))); 
	  else 
	     t=new(&tcp,port);
	  t->count++;
	  t->size+=hdr.len;
	  pthread_mutex_unlock(&mutex);
	  }
	
	break;
      default:
      }
    }
  }
   
   pcap_close(pd);
}

//this function, is called from SIGALRM, and is used to update the screen
void screen_update() {
  int i, do_ipprint, last[3]={0};
  unsigned long long sump, sums, valp, vals;
  struct node *sorted[24]={0};
  char str1[7], str2[7];
  char *strtype, strtypeb[15];
  time_t t;

  pthread_mutex_lock(&mutex);
  sump=countsump(&ether);
  sums=countsums(&ether);
  largestn(&ether,sorted,24,sort_type);
  pthread_mutex_unlock(&mutex);

  if (ui_type == 's')
    sl_header();
  if (ui_type == 't' && t_print_head)
    t_header();
  
  sprintf_nice(str1,sump);
  sprintf_nice(str2,sums);
  if(sump>0) 
    i = sums/sump;
  else
    i = 0;

  if (ui_type == 's')
    sl_print_row(100,0, str1, 100,0, str2, i, "total", 0, last);
  if (ui_type == 't')
    t_print_row(100,0, str1, 100,0, str2, i, "total", 0, last);
  
  pthread_mutex_lock(&mutex);
  //print percentages
  for(i=0; !last[0] && sorted[i]; i++) {
    //the first node in our array is there as a place holder
    //we don't want to print it out.
    if(sorted[i]->type==0xffffffff) 
      continue;

    if (i+1>=24 || !sorted[i+2] || (ui_type == 's' && y+1>maxy-1)) 
      last[0] = 1;
    

    do_ipprint = 0;
    switch(sorted[i]->type) {
    case 0x0800: strtype = "ipv4         ";
      do_ipprint = 1;                        break;
    case 0x0806: strtype = "arp          ";  break;
    case 0x6002: strtype = "dec mop      ";  break;
    case 0x6003: strtype = "dec dna-route";  break;
    case 0x6004: strtype = "dec lat      ";  break;
    case 0x8035: strtype = "rarp         ";  break;
    case 0x809b: strtype = "appletalk DDP";  break;
    case 0x8137: strtype = "novell ipx-nw";  break;
    case 0x86dd: strtype = "ipv6         ";  break;
    default:     
      sprintf(strtypeb, "0x%4.4x       ", sorted[i]->type);
      strtype = strtypeb;
    }
    valp=((long long)sorted[i]->count*10000)/sump;
    sprintf_nice(str1,sorted[i]->count);
    vals=((long long)sorted[i]->size*10000/sums);
    sprintf_nice(str2,sorted[i]->size);

    if (ui_type == 's')
      sl_print_row(valp/100, valp%100, str1, vals/100, vals%100, str2, 
                   sorted[i]->size/sorted[i]->count, strtype, 1, last);
    if (ui_type == 't')
      t_print_row(valp/100, valp%100, str1, vals/100, vals%100, str2, 
                   sorted[i]->size/sorted[i]->count, strtype, 1, last);

    if (do_ipprint)
      ipprint(last);
  }
  pthread_mutex_unlock(&mutex);
  
  if (ui_type == 's')
    SLsmg_refresh();
}

void ipprint(int last[]) {
  int j, do_tcpprint;
  struct node *sorted[24]={0};
  char str1[7], str2[7], *strtype, strtype2[14];
  unsigned long long valp, vals, sump, sums;
  sump=countsump(&ether);
  sums=countsums(&ether);

  largestn(&ip,sorted,24,sort_type);
 
  last[1] = 0;
  for(j=0; !last[1] && sorted[j]; j++) {
    if(sorted[j]->type==-1) 
      continue;
 
    if (j+1>=24 || !sorted[j+2] || (ui_type == 's' && y+1>maxy-1)) 
      last[1] = 1;

    do_tcpprint = 0;
    switch(sorted[j]->type) {
    case 00:  strtype = "reserved     "; break;
    case 01:  strtype = "icmp         "; break;
    case 02:  strtype = "igmp         "; break;
    case 06:  strtype = "tcp          ";  
              do_tcpprint = 1;           break;
    case 17:  strtype = "udp          "; break;
    case 21:  strtype = "prm          "; break;
    case 31:  strtype = "MFE NSP      "; break;
    case 34:  strtype = "3pc          "; break;
    case 70:  strtype = "gopher       "; break;
    case 113: strtype = "auth         "; break;
    default:
      sprintf(strtype2, "%d           ",sorted[j]->type);	
      strtype = strtype2; 
    }
    valp=((long long)sorted[j]->count * 10000)/sump;
    sprintf_nice(str1,sorted[j]->count);
    vals=((long long)sorted[j]->size * 10000)/sums;
    sprintf_nice(str2,sorted[j]->size);

    if (ui_type == 's')
      sl_print_row(valp/100, valp%100, str1, vals/100, vals%100, str2,
                   sorted[j]->size/sorted[j]->count, strtype, 2, last);
    if (ui_type == 't')
      t_print_row(valp/100, valp%100, str1, vals/100, vals%100, str2,
                  sorted[j]->size/sorted[j]->count, strtype, 2, last);

    if (do_tcpprint)
      tcpprint(last);
  } 
}

void tcpprint(int last[]) {
  int j;
  struct node *sorted[24]={0};
  char str1[7], str2[7], *strtype, strtype2[14];
  unsigned long long vals, valp, sums, sump;
  sump=countsump(&ether);
  sums=countsums(&ether);

  largestn(&tcp,sorted,24,sort_type);

  last[2] = 0;
  for(j=0; !last[2] && sorted[j]; j++) {
    if(sorted[j]->type==-1) 
      continue;

    if (j+1>=24 || !sorted[j+2] || (ui_type == 's' && y+1>maxy-1)) 
      last[2] = 1;

    switch(sorted[j]->type) {
    case  20: strtype = "ftp-data     "; break;
    case  21: strtype = "ftp          "; break;
    case  22: strtype = "ssh          "; break;
    case  25: strtype = "smtp         "; break;
    case  79: strtype = "finger       "; break;
    case  80: strtype = "http         "; break;
    case 110: strtype = "pop3         "; break;
    case 139: strtype = "netbios-ssn  "; break;
    case 220: strtype = "imap3        "; break;
    case 515: strtype = "printer      "; break;
    default:
      sprintf(strtype2, "%d           ",sorted[j]->type);	
      strtype = strtype2;
    }
    valp=((long long)sorted[j]->count * 10000)/sump;
    sprintf_nice(str1,sorted[j]->count);
    vals=((long long)sorted[j]->size * 10000)/sums;
    sprintf_nice(str2,sorted[j]->size);

    if (ui_type == 's')
      sl_print_row(valp/100, valp%100, str1, vals/100, vals%100, str2,
                   sorted[j]->size/sorted[j]->count, strtype, 3, last);
    if (ui_type == 't')
      t_print_row(valp/100, valp%100, str1, vals/100, vals%100, str2,
                  sorted[j]->size/sorted[j]->count, strtype, 3, last);
  } 
}

void dieing() {
  if (ui_type == 's') {
    SLsmg_reset_smg();
    SLang_reset_tty();
  }
  pcap_close(pd);
  exit(0);
}

void ui_run() {
  int i;
  start_time=time(0);

  if (ui_type == 's') {
    sl_init();
    screen_update();
  }

  for(i=0; !runcount || i<runcount; i++) {
    if (ui_type == 's')
      sl_check_kb();

    usleep(delay);
    screen_update();
  }

  dieing();
}


void change_screen_size() {
  SLtt_get_screen_size();
  SLang_init_tty(-1,0,0);
  SLsmg_reset_smg();
  SLsmg_init_smg ();

  SLsmg_cls();
  screen_update();

  signal(SIGWINCH,change_screen_size);
}

pcap_t *pcap_start(char *filterstr) {
  char *device;
  int datalink, snaplen=200;
  char errbuf[PCAP_ERRBUF_SIZE];
  unsigned int localnet, netmask;
  struct bpf_program fcode;
  pcap_t *pd;

  if(dev == 0) {
    if((device=pcap_lookupdev(errbuf))==NULL) {
      fprintf(stderr,"Can not open device: %s\n",errbuf);
      exit(1);
    }
  }
  else 
    device=dev;

  if((pd=pcap_open_live(device, snaplen, 1, 500, errbuf))==NULL) {
    fprintf(stderr,"Can not open device: %s\n",errbuf);
    if(getuid() != 0) {
      fprintf(stderr,"You may not have permissions to open this device\n");
    }
    exit(1);
  }
  
  if(pcap_lookupnet(device, &localnet, &netmask, errbuf) < 0) {
    fprintf(stderr,"pcap_open_live: %s\n",errbuf);
    exit(1);
  }
  
  if(pcap_compile(pd, &fcode, filterstr, 1, netmask) < 0) {
    fprintf(stderr,"pcap_compile: %s\n",pcap_geterr(pd));
    exit(1);
  }
  
  if(pcap_setfilter(pd,&fcode)<0) {
    fprintf(stderr,"pcap_setfilter: %s\n",pcap_geterr(pd));
    exit(1);
  }
  
  if((datalink=pcap_datalink(pd))<0) {
    fprintf(stderr,"pcap_datalink: %s\n",pcap_geterr(pd));
    exit(1);
  }
  
  return pd;
}

void sprintf_nice(char *str, unsigned int data) {
  if (data/1000==0) 
    sprintf(str,"%6.1f ",(float) data);
  else if (data/1024<1024) 
    sprintf(str,"%6.1fk",data/1024.0);
  else if (data/1048576<1024)
    sprintf(str,"%6.1fm",data/1048576.0);
  else 
    sprintf(str,"%6.1fg",data/1073741824.0);
}

void change_delay() {
  char str[10];
  unsigned int x,i=0;

  SLsmg_set_color(0);
  SLsmg_gotorc(0,0);
  SLsmg_printf("Please enter a new delay time in seconds:");
  for(x=41;x<SLtt_Screen_Cols;x++)
    SLsmg_printf(" ");
  SLsmg_refresh();
  while(i<10 && SLang_input_pending(25)) {
    x=SLang_getkey();
    switch(x) {
    case  '1':
    case  '2':
    case  '3':
    case  '4':
    case  '5':
    case  '6':
    case  '7':
    case  '8':
    case  '9':
    case  '0': 
    case  '.':
      str[i++]=x;
      SLsmg_gotorc(0,41+i);
      SLsmg_write_char(x);
      SLsmg_refresh();
      break;
    case 127: //backspace
      SLsmg_gotorc(0,41+i);
      SLsmg_write_char(' ');
      SLsmg_refresh();
      str[--i]=0;
    case '\r': 
      str[i]=0;
      delay=(1000000*atof(str));
      break;
    }
  }
}


void change_sort_type() {
  unsigned int x,i=0;

  SLsmg_set_color(0);
  SLsmg_gotorc(0,0);
  SLsmg_printf("Change sorting to [p]ackets or [s]ize:");
  for(x=41;x<SLtt_Screen_Cols;x++)
    SLsmg_printf(" ");
  SLsmg_refresh();
  while(i<10 && SLang_input_pending(25)) {
    x=SLang_getkey();
    switch(x) {
    case 'p': sort_type='p'; return;
    case 's': sort_type='s'; return;
    }
  }
}

void show_help() {
  int y=0;
  SLsmg_cls();
  SLsmg_gotorc(y++,0);
  
  SLsmg_set_color(header_c);

  SLsmg_printf("Nettop: %s", VERSION);
  SLsmg_gotorc(y++,0);
  SLsmg_printf("Written by Scott Parish, copyright 2000, BSD");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("Compiled at %s, %s",__TIME__,  __DATE__);
  y+=2;
  SLsmg_gotorc(y++,0);
  
  SLsmg_printf("  Key     Effect");
  SLsmg_set_color(0);
  SLsmg_gotorc(y++,0);
  SLsmg_printf("   c      clear the data, and start all counting over");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("   l      toggle drawing tree lines");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("   o      change the order of sort (by [p]ackets or [s]ize)");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("   s      change update delay time (default 1 sec)");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("q or x    exit nettop");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("? or h    this screen");
  SLsmg_gotorc(y++,0);
  SLsmg_printf("<space>   pause display from updating (but keep sniffing)");
  SLsmg_gotorc(y++,0);	

  SLsmg_gotorc(y++,0);
  SLsmg_printf(" %%pkts    what fraction of total packets are from this type");
  SLsmg_gotorc(y++,0);
  SLsmg_printf(" total    total number of packets from this type");
  SLsmg_gotorc(y++,0);
  SLsmg_printf(" %%size    what fraction by size of all packets are type");
  SLsmg_gotorc(y++,0);
  SLsmg_printf(" sz/pkt   What is the average size of each packet");
  SLsmg_gotorc(y++,0);
  
  SLsmg_gotorc(SLtt_Screen_Rows-2,0);
  SLsmg_printf("Press any key to return to program");
  SLsmg_refresh();
  
  SLang_getkey();
  SLsmg_cls();
}

void pause_display() {
   int x, center;
   SLsmg_gotorc(0,0);
   SLsmg_set_color(header_c);
   center=SLtt_Screen_Cols/2-8;
   for(x=0;x<center;x++)
            SLsmg_printf(" ");
   SLsmg_printf("--== paused ==--");
   for(;x<SLtt_Screen_Cols;x++) 
      SLsmg_printf(" ");
   SLsmg_refresh();
   SLang_getkey();
}

void clear_count() {
  pthread_mutex_lock(&mutex);
  ndelete(&ether);
  ndelete(&ip);
  ndelete(&tcp);
  pthread_mutex_unlock(&mutex);
  start_time = time(0);
  SLsmg_cls();
}

//
// Slang output
//

void sl_init() {
  SLtt_get_terminfo();
  SLang_init_tty(-1,0,0);
  SLsmg_init_smg();
  
  SLtt_set_color(0, "normal", "lightgray", "default");
  SLtt_set_color(total_c, "total", "yellow", "default");
  SLtt_set_color(ethernet_c, "level1", "brightcyan", "default");
  SLtt_set_color(ip_c, "level2", "brightgreen", "default");
  SLtt_set_color(tcp_c, "level3", "brightmagenta", "default");
  SLtt_set_color(bar_c, "bar", "blue", "lightgray");
  SLtt_set_color(header_c, "total", "white", "default");
  SLtt_set_color(50, "total", "gray", "default");
  SLtt_set_color(51, "total", "lightgray", "default");
  SLtt_set_color(52, "total", "white", "default");
  signal(SIGWINCH,change_screen_size);
}

void sl_check_kb() {
  while(SLang_input_pending(0)) {
    switch(SLang_getkey()) {
    case  's': change_delay();           break;
    case  'c': clear_count();		 break;
    case  'x':
    case  'q': dieing();            	 break; 
    case  '?':
    case  'h': show_help(); 	 	 break;
    case  'l': print_lines=!print_lines; break;
    case  'o': change_sort_type();  	 break;
    case  ' ': pause_display();     	 break;
    case '': SLsmg_cls();		 break;
    }
  }
}

// header is printed at the top of every update
void sl_header() {
  int x=0,a,b,c,i;
  int time_diff;
  time_t t;
  SLsmg_gotorc(0,0);
 
  x=0;
  y=1;

  //print clock
  SLsmg_set_color(header_c);
  t=time(0);
  SLsmg_printf("%s",ctime(&t));
  x=24;

  //print counter
  for(;x<SLtt_Screen_Cols-10;x++) 
     SLsmg_printf(" ");
  time_diff=time(0)-start_time;
  a=time_diff/3600;
  b=time_diff%3600/60;
  c=(time_diff%3600)%60;
  SLsmg_set_color(header_c);
  SLsmg_printf("%4d:%2.2d:%2.2d",a,b,c);
  x=0;

  SLsmg_gotorc(y++,x);
  
  x=11;
  maxy=SLtt_Screen_Rows;
  maxx=SLtt_Screen_Cols;

  //print bar
  SLsmg_set_color(bar_c);
  SLsmg_printf("  %%pkts  total    %%size   total   sz/pkt");
  for(i=x; i<maxx-43; i++)
    SLsmg_printf(" ");
  SLsmg_printf("type");
  for(i=(x+=4); i<maxx; i++)
    SLsmg_printf(" ");
  SLsmg_gotorc(y++,0);
}

void sl_print_row(int pkt, int pktf, char *str1, int tot, int totf,  
                  char *str2, int szpkt, char *type, int level, int last[]) {
  int i;
  SLsmg_set_color(level+1);
  SLsmg_printf("%3d.%2.2d%% %5s",pkt,pktf,str1);
  SLsmg_printf("  %3d.%2.2d%% %5s",tot,totf,str2);
  SLsmg_printf("  %6d", szpkt);
  SLsmg_gotorc(y-1,maxx-20);

  if (print_lines) {
    SLsmg_set_color(51);
    for (i=0; i < level-1; i++)
      if (last[i]) SLsmg_printf("  ");
      else SLsmg_printf("| ");
    if (level)
      if (last[level-1]) SLsmg_printf("`-");
      else SLsmg_printf("|-");
  }

  SLsmg_set_color(level+1);
  SLsmg_printf(type);
  SLsmg_gotorc(y++,0);
}



//
// text interface
//
void t_header() {
  printf("\n  %%pkts  total    %%size   total   sz/pkt  type\n");
  printf("-------------------------------------------------\n");
}

void t_print_row(int pkt, int pktf, char *str1, int tot, int totf,  
                 char *str2, int szpkt, char *type, int level, int last[]) {
  int i;
  printf("%3d.%2.2d%% %5s  %3d.%2.2d%% %5s  %6d  ", 
         pkt, pktf, str1, tot, totf,str2, szpkt);

  if (print_lines) {
    for (i=0; i < level-1; i++) 
      if (last[i]) printf("  ");
      else printf("| ");
    if (level) 
      if (last[level-1]) printf("`-");
      else printf("|-");
  }
  
  printf("%s\n", type);
}



//
// usage - print the command line args options
//
void usage() {
  fprintf(stderr,"Usage: nettop [-s] [-p] [-l] [-t [-h]] [-i dev] [-d delay] [-n num] [-f filter]\n");
}
